from parsomics_core.plugin_utils import PluginInitializer

from parsomics_plugin_proteinfer.populate import populate_proteinfer

initializer = PluginInitializer(
    subject="proteinfer",
    plugin_name="parsomics-plugin-proteinfer",
    populate_func=populate_proteinfer,
)



from typing import Sequence

from parsomics_core.globals.database import engine
from parsomics_core.populate import process_files
from timeit_decorator import timeit

from .processor import ProteinferOutputProcessor


@timeit()
def populate_proteinfer(
    run_info: dict, assembly_key: int, dereplicated_genomes: Sequence[str]
) -> None:
    def process_proteinfer_files(
        output_directory, assembly_key, run_key, tool_key, dereplicated_genomes
    ):
        proteinfer_output_processor = ProteinferOutputProcessor(
            output_directory=output_directory,
            assembly_key=assembly_key,
            run_key=run_key,
            tool_key=tool_key,
            dereplicated_genomes=dereplicated_genomes,
        )
        proteinfer_output_processor.process_proteinfer_files(engine)

    process_files(
        run_info,
        assembly_key,
        dereplicated_genomes,
        "proteinfer",
        process_proteinfer_files,
    )



from typing import Sequence

from parsomics_core.factories import FileFactory

from .validated_file import ProteinferValidatedFile


class ProteinferFileFactory(FileFactory):
    def __init__(self, path: str, dereplicated_genomes: Sequence[str]):
        return super().__init__(
            validation_class=ProteinferValidatedFile,
            path=path,
            dereplicated_genomes=dereplicated_genomes,
        )



import logging
from typing import Sequence

from parsomics_core.entities.files.protein_annotation.file.models import (
    ProteinAnnotationFile,
    ProteinAnnotationFileDemand,
)
from parsomics_core.entities.files.protein_annotation.file.transactions import (
    ProteinAnnotationFileTransactions,
)
from parsomics_core.processors._helpers import retrieve_genome_key
from pydantic import BaseModel
from sqlalchemy import Engine
from sqlmodel import Session

from .file_factory import ProteinferFileFactory
from .parser import ProteinferTsvParser
from .validated_file import ProteinferValidatedFile


class ProteinferOutputProcessor(BaseModel):
    output_directory: str
    dereplicated_genomes: Sequence[str]
    assembly_key: int
    run_key: int
    tool_key: int

    def process_proteinfer_files(self, engine: Engine):
        proteinfer_file_factory: ProteinferFileFactory = ProteinferFileFactory(
            self.output_directory,
            self.dereplicated_genomes,
        )

        proteinfer_files: list[ProteinferValidatedFile] = (
            proteinfer_file_factory.assemble()
        )
        for f in proteinfer_files:
            genome_key = retrieve_genome_key(engine, f, self.assembly_key)
            run_key = self.run_key

            protein_annotation_file_demand_model = ProteinAnnotationFileDemand(
                path=f.path,
                run_key=run_key,
                genome_key=genome_key,
            )

            with Session(engine) as session:
                protein_annotation_file: ProteinAnnotationFile = (
                    ProteinAnnotationFile.model_validate(
                        ProteinAnnotationFileTransactions().demand(
                            session,
                            protein_annotation_file_demand_model,
                        )
                    )
                )

            proteinfer_parser = ProteinferTsvParser(
                file=protein_annotation_file,
                assembly_key=self.assembly_key,
                tool_key=self.tool_key,
            )
            proteinfer_parser.parse(engine)

        logging.info(
            f"Finished adding all ProteInfer files on {self.output_directory} to the database."
        )



from pathlib import Path
from typing import ClassVar

from parsomics_core.entities.files.validated_file import ValidatedFileWithGenome


class ProteinferValidatedFile(ValidatedFileWithGenome):
    _VALID_FILE_TERMINATIONS: ClassVar[list[str]] = ["_ProteInfer_out.tsv", ".tsv"]

    @property
    def genome_name(self) -> str:
        path_obj = Path(self.path)
        return path_obj.name.replace("_ProteInfer_out.tsv", "").replace(".tsv", "")



import logging

import polars as pl
from parsomics_core.entities import ProteinAnnotationEntry, ProteinAnnotationFile
from parsomics_core.entities.workflow.source import (
    Source,
    SourceCreate,
    SourceTransactions,
)
from parsomics_core.plugin_utils import search_protein_by_name
from pydantic import BaseModel
from sqlalchemy.exc import IntegrityError
from sqlmodel import Session

from parsomics_plugin_proteinfer.annotation_type import ProteinferAnnotationType


class ProteinferTsvParser(BaseModel):
    file: ProteinAnnotationFile
    assembly_key: int
    tool_key: int

    def to_dataframe(self) -> pl.DataFrame:
        df = pl.read_csv(
            self.file.path,
            separator="\t",
            null_values=["-"],
        )

        df = df.rename(
            {
                "sequence_name": "protein_name",
                "predicted_label": "accession",
                "confidence": "score",
            }
        )

        df = df.filter(pl.col("accession").is_not_null())
        df = self._remove_go_annotations_from_df(df)

        df = df.with_columns(
            pl.when(pl.col("accession").str.contains("Pfam"))
            .then(pl.lit("PFAM"))
            .otherwise(pl.lit("EC_NUMBER"))
            .alias("annotation_type")
        )

        df = df.with_columns(
            pl.when(pl.col("accession").str.contains(r"^(Pfam:|EC:)"))
            .then(pl.col("accession").str.replace(r"^(Pfam:|EC:)", ""))
            .otherwise(pl.col("accession"))
            .alias("accession")
        )

        return df

    def _remove_go_annotations_from_df(self, df: pl.DataFrame) -> pl.DataFrame:
        df2 = df.filter(~pl.col("accession").str.contains(r"GO:\d+"))

        return df2

    def _add_file_key_to_df(self, df):
        return df.with_columns(pl.lit(self.file.key).alias("file_key"))

    def _add_protein_key_to_mappings(self, mappings):
        protein_name_to_key = {}
        for mapping in mappings:
            protein_name = mapping["protein_name"]
            if protein_name not in protein_name_to_key:
                protein_key = search_protein_by_name(protein_name, self.assembly_key)
                protein_name_to_key[protein_name] = protein_key

            protein_key = protein_name_to_key[protein_name]
            mapping["protein_key"] = protein_key
            mapping.pop("protein_name")

    def _add_empty_details(self, mappings):
        for mapping in mappings:
            mapping["details"] = {}

    def parse(self, engine) -> None:
        df = self.to_dataframe()
        df = self._add_file_key_to_df(df)
        mappings = df.to_dicts()
        self._add_protein_key_to_mappings(mappings)
        self._add_empty_details(mappings)

        for mapping in mappings:
            mapping["annotation_type"] = ProteinferAnnotationType(
                mapping["annotation_type"]
            )

        with Session(engine) as session:
            try:
                session.bulk_insert_mappings(ProteinAnnotationEntry, mappings)
                session.commit()
                logging.info(
                    f"Added ProteInfer entries from {self.file.path} to the database."
                )
            except IntegrityError as e:
                logging.warning(
                    f"Failed to add ProteInfer entries from {self.file.path} to "
                    f"the database. Exception caught: {e}"
                )
